from .base import BaseDataSource


class TestingDataSource(BaseDataSource):
    """
    A data source that keeps all data in RAM, and starts off with 0s.

    Analog inputs and discrete inputs are always 0.
    """

    def __init__(self):
        self.holding_registers = {}
        self.coils = {}

    def get_analog_input(self, unit_id: int, address: int) -> int:
        return 0

    def get_discrete_input(self, unit_id: int, address: int) -> bool:
        return False

    def get_holding_register(self, unit_id: int, address: int) -> int:
        return self.holding_registers.get((unit_id, address), 0)

    def get_coil(self, unit_id: int, address: int) -> bool:
        return self.coils.get((unit_id, address), False)

    def set_holding_register(self, unit_id: int, address: int, value: int) -> None:
        self.holding_registers[(unit_id, address)] = value

    def set_coil(self, unit_id: int, address: int, value: bool) -> None:
        self.coils[(unit_id, address)] = value



import struct
import weakref

from satella.coding import rethrow_as

from modbus_tcp_server.data_source import BaseDataSource
from modbus_tcp_server.datagrams import MODBUSTCPMessage
from .bits import BitStream, BitConsumer
from .exceptions import InvalidFrame, \
    CustomMODBUSError

STRUCT_HH = struct.Struct('>HH')
STRUCT_HHB = struct.Struct('>HHB')


def read_holding_registers(db: BaseDataSource, unit_id: int, address_start: int, amount: int):
    output = []
    for address in range(address_start, address_start + amount):
        output.append(db.get_holding_register(unit_id, address))
    output_len = len(output)
    return struct.pack('>H' + ('H' * output_len), output_len * 2, *output)


def read_analog_inputs(db: BaseDataSource, unit_id: int, address_start: int, amount: int):
    output = []
    for address in range(address_start, address_start + amount):
        output.append(db.get_analog_input(unit_id, address))
    output_len = len(output)
    return struct.pack('>H' + ('H' * output_len), output_len * 2, *output)


def read_coil(db: BaseDataSource, unit_id: int, address_start: int, amount: int):
    bs = BitStream()
    for address in range(address_start, address_start + amount):
        bs.add(db.get_coil(unit_id, address))
    return bytes([len(bs)]) + bytes(bs)


def read_discrete_input(db: BaseDataSource, unit_id: int, address_start: int, amount: int):
    bs = BitStream()
    for address in range(address_start, address_start + amount):
        bs.add(db.get_discrete_input(unit_id, address))
    return bytes([len(bs)]) + bytes(bs)


def write_single_coil(db: BaseDataSource, unit_id: int, address: int, value: int) -> bytes:
    db.set_coil(unit_id, address, value == 0xFF00)
    return STRUCT_HH.pack(address, value)


def write_single_register(db: BaseDataSource, unit_id: int, address: int, value: int) -> bytes:
    db.set_holding_register(unit_id, address, value)
    return STRUCT_HH.pack(address, value)


def write_multiple_registers(db: BaseDataSource, unit_id: int, msg: MODBUSTCPMessage) -> bytes:
    address, amount, databytes, reg_data = STRUCT_HHB.unpack(msg.data[1:6]), msg.data[6:]
    if databytes != 2 * amount:
        raise InvalidFrame('Mismatch between writing amount and no of bytes')

    with rethrow_as(struct.error, InvalidFrame):
        for address, value in zip(range(address, address + amount),
                                  struct.unpack('>' + ('H' * amount), reg_data)):
            db.set_holding_register(unit_id, address, value)
    return STRUCT_HH.pack(address, amount)


def write_multiple_coils(db: BaseDataSource, unit_id: int, msg: MODBUSTCPMessage) -> bytes:
    address, amount, databytes = STRUCT_HHB.unpack(msg.data[1:6])
    target_db = amount // 8
    if amount % 8:
        target_db += 1
    if target_db != databytes:
        raise InvalidFrame('Mismatch between writing amount and no of bytes')
    stream = BitConsumer(msg.data[6:])
    for address, value in zip(range(address, address + amount), stream):
        db.set_coil(unit_id, address, value)
    return STRUCT_HH.pack(address, amount)


TRANSLATION_TABLE = {
    0x03: (read_holding_registers, STRUCT_HH),
    0x04: (read_analog_inputs, STRUCT_HH),
    0x01: (read_coil, STRUCT_HH),
    0x02: (read_discrete_input, STRUCT_HH),
    0x05: (write_single_coil, STRUCT_HH),
    0x06: (write_single_register, STRUCT_HH),
    0x10: (write_multiple_registers, None),
    0x0F: (write_multiple_coils, None)
}


class ModbusProcessor:
    __slots__ = ('server', 'struct_cache')

    def __init__(self, server):
        self.server = weakref.proxy(server)

    @property
    def data_source(self) -> BaseDataSource:
        return self.server.data_source

    def process(self, msg: MODBUSTCPMessage) -> MODBUSTCPMessage:
        """
        :raises InvalidFrame: invalid data frame
        """
        try:
            proc_fun, struct_inst = TRANSLATION_TABLE[msg.data[0]]
            if struct_inst is not None:
                args = struct_inst.unpack(msg.data[1:])
                b = proc_fun(self.data_source, msg.unit_id, *args)
            else:
                b = proc_fun(self.data_source, msg.unit_id, msg)
            return msg.respond(b)
        except KeyError:
            return msg.respond(b'\x01', True)
        except CustomMODBUSError as e:
            return msg.respond(bytes([e.error_code]), True)



import logging
import socket
import time

from satella.coding import silence_excs
from satella.coding.concurrent import TerminableThread

from ..datagrams import MODBUSTCPMessage
from ..exceptions import InvalidFrame

logger = logging.getLogger(__name__)

MAX_TIME_WITHOUT_ACTIVITY = 60.0    # 60 seconds


class ConnectionThread(TerminableThread):
    def __init__(self, sock, addr, server):
        super().__init__(terminate_on=(socket.error, InvalidFrame), daemon=True)
        self.socket = sock
        self.server = server
        self.addr = addr
        self.socket.setblocking(True)
        self.socket.settimeout(5)
        self.buffer = bytearray()
        self.last_activity = time.monotonic()

    def cleanup(self):
        self.socket.close()

    @silence_excs(socket.timeout)
    def loop(self):
        # Check for termination conditions
        # noinspection PyProtectedMember
        if self.server._terminating:
            raise socket.error()
        if time.monotonic() - self.last_activity > MAX_TIME_WITHOUT_ACTIVITY:
            raise socket.error()

        # Get the data
        data = self.socket.recv(128)
        if not data:
            raise socket.error()
        logger.info('Received %s', repr(data))
        self.buffer.extend(data)

        # Extract MODBUS packets
        while self.buffer:
            try:
                packet = MODBUSTCPMessage.from_bytes(self.buffer)
            except ValueError:
                break
            del self.buffer[:len(packet)]
            msg = self.server.process_message(packet)
            b = bytes(msg)
            logger.debug('Sent %s', repr(b))
            try:
                self.socket.sendall(b)
            except socket.timeout:
                # Timeout on sendall means the client is extremely slow
                raise socket.error()
            self.last_activity = time.monotonic()



class BitStream:
    def __init__(self):
        self.pointer = 0
        self.data = bytearray()

    def __len__(self):
        return len(self.data)

    def __bytes__(self):
        return bytes(self.data)

    def add(self, bit: bool):
        ptr = self.pointer % 8
        value = 1 << ptr if bit else 0
        if not ptr:
            self.data += b'\x00'
        self.data[-1] = self.data[-1] | value
        self.pointer += 1


class BitConsumer:
    def __init__(self, data: bytes):
        self.data = bytearray(data)
        self.pointer = 0

    def __iter__(self):
        return self

    def __next__(self) -> bool:
        if not self.data:
            raise StopIteration()
        if self.pointer == 8:
            del self.data[0]
            self.pointer = 0
        try:
            return (self.data[0] >> self.pointer) & 1
        finally:
            self.pointer += 1



import logging
import sys

from satella.os import hang_until_sig

from modbus_tcp_server.network import ModbusTCPServer


def run():
    """Entry point for the MODBUS TCP server"""
    if len(sys.argv) < 2 or '-h' in sys.argv or '--help' in sys.argv or 'help' in sys.argv:
        print('''Usage:
        
modbus-tcp-server <name of interface to bind to> <optional port to bind to>

    Extra allowed arguments (after the name of interface and optional port)
        -v : display inbound traffic
        -vv : display inbound and outbound traffic

    0.0.0.0 will bind to all interfaces
    127.0.0.1 will bind to localhost only, which means that modbus-tcp-server will be
    reachable only from this machine
    
    Default port is 502
''')
        sys.exit(0)

    if len(sys.argv) < 3:
        port = 502
    else:
        port = int(sys.argv[2])

    host = sys.argv[1]

    l_level = logging.WARNING
    if '-vv' in sys.argv:
        l_level = logging.DEBUG
    elif '-v' in sys.argv:
        l_level = logging.INFO

    logging.basicConfig(level=l_level)

    at = ModbusTCPServer(host, port).start()
    hang_until_sig()
    at.terminate().join()
    sys.exit(0)



import socket
import typing as tp

from satella.coding import silence_excs
from satella.coding.concurrent import TerminableThread

from .conn_thread import ConnectionThread
from ..data_source import BaseDataSource, TestingDataSource
from ..datagrams import MODBUSTCPMessage
from ..processor import ModbusProcessor


class ModbusTCPServer(TerminableThread):
    def __init__(self, bind_ifc: str, bind_port: int,
                 data_source: tp.Optional[BaseDataSource] = None,
                 backlog: int = 128):
        super().__init__()
        if data_source is None:
            data_source = TestingDataSource()
        self.data_source = data_source
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.bind((bind_ifc, bind_port))
        self.backlog = backlog
        self.processor = ModbusProcessor(self)

    def prepare(self) -> None:
        self.socket.listen(self.backlog)
        self.socket.setblocking(True)
        self.socket.settimeout(5)

    def process_message(self, msg: MODBUSTCPMessage) -> MODBUSTCPMessage:
        return self.processor.process(msg)

    def cleanup(self):
        self.socket.close()

    @silence_excs(socket.timeout)
    def loop(self) -> None:
        sock, addr = self.socket.accept()
        ConnectionThread(sock, addr, self).start()



import struct

from satella.coding import rethrow_as

from .exceptions import InvalidFrame

STRUCT_HDR = struct.Struct('>HHHB')
SHDR_SIZE = STRUCT_HDR.size


class MODBUSTCPMessage:
    def __init__(self, tid: int, pid: int, unit_id: int, data: bytes):
        self.tid = tid
        self.pid = pid
        self.unit_id = unit_id
        self.data = data

    def __bytes__(self):
        return STRUCT_HDR.pack(self.tid, self.pid, len(self.data) + 1, self.unit_id) + self.data

    def __len__(self) -> int:
        return SHDR_SIZE + len(self.data)

    @classmethod
    @rethrow_as((IndexError, struct.error), ValueError)
    def from_bytes(cls, d: bytes):
        tid, pid, length, uid = STRUCT_HDR.unpack(d[:SHDR_SIZE])
        if pid != 0x00:
            raise InvalidFrame('Protocol ID not 0')
        data = d[SHDR_SIZE:SHDR_SIZE + length - 1]
        if len(data) < length - 1:
            raise ValueError('Input data too short')
        return MODBUSTCPMessage(tid, pid, uid, data)

    def respond(self, data: bytes, exception: bool = False) -> 'MODBUSTCPMessage':
        """
        Provide a response to this datagram

        :param data: data to respond with. This will be appended to the function code
            of this datagram!
        :param exception: is this an exceptional response?
        :return: a MODBUSTCPMessage
        """
        if exception:
            return MODBUSTCPMessage(self.tid, 0, self.unit_id, bytes([self.data[0] & 0x80]) + data)
        else:
            return MODBUSTCPMessage(self.tid, 0, self.unit_id, self.data[0:1] + data)



class ModbusTCPError(Exception):
    """Base class"""


class InvalidFrame(ModbusTCPError):
    """Frame was invalid"""


class CustomMODBUSError(ModbusTCPError):
    """
    A custom error response is to be provided

    :param error_code: this code will be returned to the user
    """

    def __init__(self, error_code: int):
        assert 0 >= error_code >= 255, 'Invalid error code'
        self.error_code = error_code


class IllegalAddress(CustomMODBUSError):
    """The address was illegal"""

    def __init__(self):
        super().__init__(0x02)


class IllegalValue(CustomMODBUSError):
    """The value was illegal"""

    def __init__(self):
        super().__init__(0x03)


class GatewayTargetDeviceFailedToRespond(CustomMODBUSError):
    """Gateway target device failed to respond"""

    def __init__(self):
        super().__init__(0x0B)


class GatewayPathUnavailable(CustomMODBUSError):
    """Gateway path unavailable"""

    def __init__(self):
        super().__init__(0x0A)



from abc import ABCMeta, abstractmethod


class BaseDataSource(metaclass=ABCMeta):
    """
    A base class for all data sources
    """

    @abstractmethod
    def get_holding_register(self, unit_id: int, address: int) -> int:
        """
        :raises IllegalAddress: address was illegal
        :raises IllegalValue: value was illegal
        :raises GatewayTargetDeviceFailedToRespond: gateway target device failed to respond
        :raises GatewayPathUnavailable: gateway path unavailable
        :raises CustomMODBUSError: custom modbus error
        :return: an int 0 >= x >= 65535
        """

    @abstractmethod
    def get_analog_input(self, unit_id: int, address: int) -> int:
        """
        :raises IllegalAddress: address was illegal
        :raises IllegalValue: value was illegal
        :raises GatewayTargetDeviceFailedToRespond: gateway target device failed to respond
        :raises GatewayPathUnavailable: gateway path unavailable
        :raises CustomMODBUSError: custom modbus error
        :return: an int 0 >= x >= 65535
        """

    @abstractmethod
    def get_discrete_input(self, unit_id: int, address: int) -> bool:
        """
        :raises IllegalAddress: address was illegal
        :raises IllegalValue: value was illegal
        :raises GatewayTargetDeviceFailedToRespond: gateway target device failed to respond
        :raises GatewayPathUnavailable: gateway path unavailable
        :raises CustomMODBUSError: custom modbus error
        """

    @abstractmethod
    def get_coil(self, unit_id: int, address: int) -> bool:
        """
        :raises IllegalAddress: address was illegal
        :raises IllegalValue: value was illegal
        :raises GatewayTargetDeviceFailedToRespond: gateway target device failed to respond
        :raises GatewayPathUnavailable: gateway path unavailable
        :raises CustomMODBUSError: custom modbus error
        """

    @abstractmethod
    def set_holding_register(self, unit_id: int, address: int, value: int) -> None:
        """
        :param unit_id: Unit ID
        :param address: address to write
        :param value: an int 0 >= x >= 65535
        :raises IllegalAddress: address was illegal
        :raises IllegalValue: value was illegal
        :raises GatewayTargetDeviceFailedToRespond: gateway target device failed to respond
        :raises GatewayPathUnavailable: gateway path unavailable
        :raises CustomMODBUSError: custom modbus error
        """

    @abstractmethod
    def set_coil(self, unit_id: int, address: int, value: bool) -> None:
        """
        :raises IllegalAddress: address was illegal
        :raises IllegalValue: value was illegal
        :raises GatewayTargetDeviceFailedToRespond: gateway target device failed to respond
        :raises GatewayPathUnavailable: gateway path unavailable
        :raises CustomMODBUSError: custom modbus error
        """



from setuptools import find_packages
from distutils.core import setup
from modbus_tcp_server import __version__


setup(version=__version__,
      packages=find_packages(include=['modbus_tcp_server', 'modbus_tcp_server.*']),
      install_requires=['satella'],
      python_requires='!=2.7.*,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*,!=3.5.*',
      zip_safe=True,
      entry_points={
            'console_scripts': [
                  'modbus-tcp-server = modbus_tcp_server.run:run'
            ]
      }

      )
